#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <grp.h>
#include <sys/types.h>
#include <pwd.h>

/* Maximum width of the terminal */
#define MAX_WIDTH 	134
#define VERSION_MAJOR	0
#define VERSION_MINOR	9
#define VERSION_RELEASE	"beta"

void print_line(char *msg);
void print_version(void);
void print_usage(void);
void print_array(int lines, char *msg[]);
void create_line(char *output, char *filename, int max_fname_size, struct stat *st);
char *get_gid(int gid);
char *get_uid(int uid);
char *nice_size(long size);
char *nice_time(time_t t);

int watch_size=0, watch_user=0, watch_group=0, watch_access=0, watch_modified=0, watch_changed=0;
int update_freq=1, whip_me_beat_me_make_me_y2k_compliant=0;
int dont_follow_symlinks=0;

char **files;
char **output;

int main(int argc, char *argv[])
{
	int i,count,maxfsize=0,l,rv;
	int watch_nodef=0;
	struct stat st;

	while((i = getopt(argc, argv, "VsugamcF:YSh")) != -1) {
		switch(i) {
		case 'V':
			print_version();
			return 0;
		case 'h':
			print_usage();
			return 0;
		case 'F':
			update_freq = atoi(optarg);
			if (update_freq < 1) {
				puts("Invalid update frequency");
				return 3;
			}
			break;
		case 'S': dont_follow_symlinks = 1; break;
		case 'Y': whip_me_beat_me_make_me_y2k_compliant = 1; break;
		case 's': watch_size = 1; watch_nodef = 1; break;
		case 'u': watch_user = 1; watch_nodef = 1; break;
		case 'g': watch_group = 1; watch_nodef = 1; break;
		case 'a': watch_access = 1; watch_nodef = 1; break;
		case 'm': watch_modified = 1; watch_nodef = 1; break;
		case 'c': watch_changed = 1; watch_nodef = 1; break;
		}
	}
	if (watch_nodef != 1) { /* Use defaults */
		watch_size = 1;
	}

	count = argc - optind;
	if (count < 1) {
		print_usage();
		return 0;
	}

	files = (char **) malloc(sizeof(char *) * count);
	output = (char **) malloc(sizeof(char *) * count);

	for (i = optind; i < argc; i++) {
		files[i-optind] = strdup(argv[i]);
		l = strlen(argv[i]);

		if (l > maxfsize)
			maxfsize = l;

		output[i-optind] = (char *) malloc(sizeof(char) * (MAX_WIDTH+1));
		putc('\n',stdout); /* May aswell move the screen down now too */ 
	}

	while(1) {
		for (i = 0; i < count; i++) {
			if (dont_follow_symlinks)
				rv = lstat(files[i],&st);
			else
				rv = stat(files[i],&st);
		
			if (rv == -1) {
				strncpy(output[i],"Error stat(2)'ing file.",MAX_WIDTH);
				continue;
			}
			create_line(output[i],files[i],maxfsize,&st);
		}
		print_array(count,output);
		sleep(update_freq);
	}
	
	return 0;
}


void create_line(char *output, char *filename, int max_fname_size, struct stat *st)
{
	char tempstr[MAX_WIDTH+1];
#define DELIM	" \t"
	sprintf(tempstr,"%-*s: ",max_fname_size,filename);
	//strcpy(tempstr,filename);
	

	if (watch_size) {
		if (S_ISDIR(st->st_mode))
			strcat(tempstr,"<dir>");
		else
			strcat(tempstr,nice_size(st->st_size));

		strcat(tempstr,DELIM);
	}
	if (watch_user) {
		strcat(tempstr,get_uid(st->st_uid));
		if (watch_group)
			strcat(tempstr,":");
		else
			strcat(tempstr,DELIM);
	}
	if (watch_group) {
		strcat(tempstr,get_gid(st->st_gid));
		strcat(tempstr,DELIM);
	}
	if (watch_access) {
		strcat(tempstr,nice_time(st->st_atime));
		strcat(tempstr,DELIM);
	}
	if (watch_modified) {
		strcat(tempstr,nice_time(st->st_mtime));
		strcat(tempstr,DELIM);
	}
	if (watch_changed) {
		strcat(tempstr,nice_time(st->st_ctime));
		strcat(tempstr,DELIM);
	}
	strcpy(output,tempstr);
}


char *nice_size(long size)
{
	static char buffer[21];
	char *n_nice;
	double n_bytes=size;

#define ONEMEG 1048576
#define ONEGIG (ONEMEG * 1024)
	if (n_bytes > ONEGIG) {
		n_bytes /= ONEGIG;
		n_nice = "GB";
	} else if (n_bytes > ONEMEG) {
		n_bytes /= ONEMEG;
		n_nice = "MB";
	} else if (n_bytes > 1024) {
		n_bytes /= 1024;
		n_nice = "KB";
	} else {
		snprintf(buffer,99,"%luB", size);
		return buffer;
	}
	snprintf(buffer,20,"%.2f%s",n_bytes,n_nice);
	return buffer;
}

char *nice_time(time_t t)
{
	struct tm *tinfo;
	static char buffer[21];

	tinfo = localtime(&t);

	/* Something to keep the paranoid happy */
	if (whip_me_beat_me_make_me_y2k_compliant) 
		snprintf(buffer,20,"%2d/%0.2d/%0.2d %0.2d:%0.2d:%0.4d",
			tinfo->tm_mday,tinfo->tm_mon+1,(tinfo->tm_year + 1900),
			tinfo->tm_hour,tinfo->tm_min,tinfo->tm_sec);
	else 
		snprintf(buffer,20,"%2d/%0.2d/%0.2d %0.2d:%0.2d:%0.2d",
			tinfo->tm_mday,tinfo->tm_mon+1,(tinfo->tm_year % 100),
			tinfo->tm_hour,tinfo->tm_min,tinfo->tm_sec);
	return buffer;
}

char *get_uid(int uid) 
{
	struct passwd *pw;
	static char buffer[21];

	pw = getpwuid(uid);
	if (pw == NULL) 
		snprintf(buffer,20,"%d",uid);
	else
		snprintf(buffer,20,"%s",pw->pw_name);
	return buffer;
}

char *get_gid(int gid) 
{
	struct group *gp;
	static char buffer[21];

	gp = getgrgid(gid);
	if (gp == NULL) 
		snprintf(buffer,20,"%d",gid);
	else
		snprintf(buffer,20,"%s",gp->gr_name);
	return buffer;
}

/* Prints this array on the screen by first going up N lines, then
 * overwritting whats currently on those lines with the contents of
 * the array (wipes to End of line). N is the number of lines in the
 * array.
 */
void print_array(int lines, char *msg[])
{
	int i;

	printf("\033[%dA",lines); /* move up N lines */
	printf("\033[G"); /* move to beginning of line */
	for(i = 0; i < lines; i++) {
		printf("\033[J"); /* Clear to EOL */
		printf(msg[i]);
		printf("\n");
	}

}

/* Prints the line, on the current line, erasing whatever else is on it */
void print_line(char *msg)
{
	printf("\033[G"); /* Back to beginning of line */
	printf(msg);
	printf("\033[J"); /* Clear to end of line (?) */
}

void print_usage()
{
	puts("watchfile [-Vsugamch] [-F <num>] <files...>\n");
	puts("You may specify as many files as you wish, but if you specify");
	puts("more than the number of lines on your terminal, results are");
	puts("unspecified. Also, the more options you include, the wider");
	puts("the output. If the output is wider than your terminal, again,");
	puts("the results are unspecified.");
	puts("Options:");
	puts("   -V      Displays the program version");
	puts("   -F <num>  Update the display every <num> seconds");
	puts("             Defaults to 1 second");
	puts("   -h      Display this screen");
	puts("   -S      Don't follow symlinks, stat the actual link file");
	puts("   -Y      Display dates with 4 digit year instead of 2");
	puts("   -s      Display file sizes");
	puts("   -u      Display file user owner");
	puts("   -g      Display file group owner");
	puts("   -a      Display last accessed time");
	puts("   -m      Display last modified[1] time");
	puts("   -c      Display last changed[2] time");
	puts("");
	puts("[1] Modified time is the time the file was last written to.");
	puts("[2] Changed time is the time the file info (eg uid, mode, etc");
	puts("    was last changed");
	puts("");
}

void print_version()
{
	printf("watchfile version %d.%d - %s\n",
		VERSION_MAJOR,VERSION_MINOR,VERSION_RELEASE);
	puts("Copyright 1999, Nick Clifford <zaf@nrc.co.nz>\n");
	puts("This file is licensed under the GNU General Public License.");
	puts("For more information on the licensing conditions and on the");
	puts("GPL, read the LICENSE file distributed with this application");
	puts("or visit this apps website.");
	puts("watchfile. Created by Nick 'Zaf' Clifford");
	puts("email: zaf@nrc.co.nz");
	puts("website: http://www.nrc.co.nz/Zaf/apps/watchfile");
}

